﻿//================================================================================
//  
//        APB3.cs        ＡＰＢ－３基板制御クラス
//  
//        (c)2011/12/01  by OjisanKoubou
//  
//================================================================================
//  
//  このプログラムは Ojisankoubou( OGAWA Ichiro ) が著作権を持っています。
//  個人の研究目的には、修正、改造など自由に使用できます。
//  発表される際は著作権表示部分を消さないようにしてください。
//  このソフトウェアの全部もしくは一部を製品に組み込んだり販売したりするなど
//  商業的利用はおやめください。
//

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using Microsoft.Win32.SafeHandles; 
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using FTD2XX_NET;

namespace APB_3 {
    public static class Apb3 {
        public static MainForm mainf;                      // 
        
        private static Mutex mutex_apb3 = new Mutex(); // 基板使用中を示すローカルミューテックス
        public static bool isft232hopen = false;       // FT232H がオープンしているか
        public static bool isconfigured = false;       // FPGA がコンフィグ済みか
        
        public static double clockfreq  = Fpga.clockfreq;   // 100MHz 周波数補正することを考慮し必ずこれを使う
        public static double nyquistfreq= 100e6 / 2;        // 
        public static int buffsize      = Fpga.buffsize;
        public static int nFFT          = Fpga.buffsize/4;  // FFT データ数
        public static double ddsdeltaf  = clockfreq / Math.Pow(2, 32);  // DDS(32bit) の周波数分解能
        
        public static byte[] tmpbuf     = new byte[4+256];
        
        public static Byte[] filebuf    = new Byte[500000]; // ファイル読み込みデータ格納用配列、本来は334kBあればOK
        public static int filesize      = 0;                // ファイルサイズ
        
        public static Stopwatch sw      = new Stopwatch();
        
        public static int adccaldly= -1;                // ADC1610S の IDELAY2 設定値
        
        
        
        // 
        // ＡＰＢ－３基板との通信を開始する
        // このクラスのほかのメソッドを使う前に明示的にこの関数を呼んで排他的使用権を得ること
        // 
        public static bool Open() {
            int err = TryOpen();
            if( err < 0 )
            {
                error.ErrorMessageBox( err );
                return false;
            }
            return true;
        }
        
        public static int TryOpen() {
            try
            {
                if(!mutex_apb3.WaitOne(TimeSpan.Zero, false)) {
//                    error.ErrorMessageBox(error.usb_mutex);
                    mainf.LogPrint("\nApb3.TryOpen: ERROR  can't get mutex");
                    return error.usb_mutex;
                }
                if (!Ft232h.FtOpen()) {
                    mutex_apb3.ReleaseMutex(); // 
                    return error.usb_connect;
                }
                if (!isconfigured) {
                    if( !DeviceInit() ) {
                        isconfigured = false;
                        mutex_apb3.ReleaseMutex();
                        return error.fpga_config;
                    }
                }
                
                Ft232h.FtPurge(); // RX/TXバッファをクリアする
                
                return 0;
            }
            catch (Exception ex)
            {
                mainf.LogPrint("\nApb3.Open: Exception ex=" + ex.ToString());
                return error.err;
            }
        }
        
        
        // 
        // ＡＰＢ－３基板との通信終了、使用権を返す
        // 
        public static void Close() {
            try {
                mutex_apb3.ReleaseMutex();
                if( isft232hopen ) Ft232h.FtPurge(); // RX/TXバッファをクリアする
            }
            catch (Exception e) {
                mainf.LogPrint("\nApb3.Close: Exception ex=" + e.Message);
            }
        }
        
        
        // 
        // ＡＰＢ－３基板との接続切断
        // 
        public static void CloseConnection() {
            mainf.LogPrint("\nApb3.CloseConnection:");
            Apb3.Ft232h.FtClose();    // 
            isconfigured = false;
            adccaldly= -1;
        }
        
        
        
        //
        //  コンフィグファイルをオープンしバッファーに読み込む
        //
        public static void OpenConfigFile()
        {
            OpenFile("Fpga Config", "bitファイル(*.bit;*.bin)|*.bit;*.bin|すべてのファイル(*.*)|*.*");
        }
        
        
        //
        //  ファイルオープンしバッファーに読み込む
        //
        public static void OpenFile( string title, string filter )
        {
            for (int i = 0; i < filebuf.Length; i++) filebuf[i] = 0;
            
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.InitialDirectory = "";
            ofd.Filter = filter;
            ofd.FilterIndex = 1;
            ofd.Title = title + " ファイルを選択してください";
            ofd.RestoreDirectory = true;
            
            filesize = 0;
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                FileStream fs = new FileStream(ofd.FileName, FileMode.Open);
                filesize = (int)fs.Length;         // ファイルサイズ
                mainf.LogPrint("\n" + ofd.FileName + "  size=" + filesize.ToString());
                fs.Read(filebuf, 0, (int)filesize);     // 全部読み込む
                fs.Close();
            }
        }
        
        
        
        // 周辺デバイスの初期化 ＆ ＦＰＧＡ Ｃｏｎｆｉｇ
        public static bool DeviceInit() {
            if(!Fpga.FpgaConfig()) return false;
            mainf.LogPrint("\nFPGA version="+FpgaSignature());
            Apb3.uda1345_Init();                        // UDA1345 初期設定
            if(!Adc1610.AdcInit()) return false;        // ＡＤＣ (ADC1610S) 初期設定
            return true;
        }
        
        
        // ＦＰＧＡ Ｓｉｇｎａｔｕｒｅ　読み出し 
        public static string FpgaSignature() {
            byte[] Rbuf = new byte[16];
            Apb3.Fpga.FpgaRead( 0, 16, Rbuf );
            if( (Rbuf[0] < 0x20) || (Rbuf[0] > 127) ) { // 印刷可能文字ではないときは１６進で表示
                string str="";
                for( int i=0; i<16; i++ )  str += Rbuf[i].ToString("X2");
                return str;
            }
            return Encoding.ASCII.GetString(Rbuf);
        }
        
        
        
        
        // ＤＤＳ０　周波数 設定        周波数変換、ＴＧ出力
        public static void dds0( double freq ) {
            int f = (int)(freq / clockfreq * Math.Pow(2, 32) + 0.5);
            Fpga.FpgaWrite( Fpga.dds0, 4, f );              // 周波数
        }
        // ＤＤＳ０　周波数 設定        周波数変換、ＴＧ出力
        public static void dds0( int f ) {
            Fpga.FpgaWrite( Fpga.dds0, 4, f );              // 周波数
        }
        
        // ＤＤＳ１　周波数 設定        信号発生器
        public static void dds1( double freq ) {
            int f = (int)(freq / clockfreq * Math.Pow(2, 32) + 0.5);
            Fpga.FpgaWrite( Fpga.dds1, 4, f );   // 周波数
        }
        // ＤＤＳ１　周波数 設定        信号発生器
        public static void dds1( int f ) {
            Fpga.FpgaWrite( Fpga.dds1, 4, f );   // 周波数
        }
        
        // ＤＤＳ２（CORDIC 低周波）　周波数、波形種類、初期位相 設定
        public static void dds2( double freq ) {
            dds2( freq, 0 );
        }
        public static void dds2( double freq, int wave ) {
            int f = (int)(freq / ddsdeltaf + 0.5);
            Fpga.FpgaWrite( Fpga.dds2, 4, f );        // 周波数
            Fpga.FpgaWrite( Fpga.dds2+4, 1, wave );   // 波形
        }
        
        // ＤＤＳ３（CORDIC 低周波）　周波数、波形種類、初期位相 設定
        public static void dds3( double freq ) {
            dds3( freq, 0, 0 );
        }
        public static void dds3( double freq, int wave ) {
            dds3( freq, wave, 0 );
        }
        public static void dds3( double freq, int wave, double phase ) {
            int f = (int)(freq / ddsdeltaf + 0.5);
            Fpga.FpgaWrite( Fpga.dds3, 4, f );        // 周波数
            Fpga.FpgaWrite( Fpga.dds3+4, 1, wave );   // 波形
            int p = (int)(phase / 180 * 32768);
            Fpga.FpgaWrite( Fpga.dds3+5, 2, p );      // 位相
        }
        
        
        // ＤＤＳ　その他設定
        public static void dds_setting( int n, int setting ) {
            int adrs;
            switch( n ) {
                case 1: adrs = Fpga.dds1+4; break;
                case 2: adrs = Fpga.dds2+4; break;
                case 3: adrs = Fpga.dds3+4; break;
                default: return;
            }
            Fpga.FpgaWrite( adrs, 4, setting );  //
        }
        
        
        
        
        
        // メモリー書き込み開始
        public static void MemwriteStart() {
            Fpga.FpgaWrite(Fpga.memcontrol, 1);
        }
        
        // メモリー書き込みステータスを読む
        public static int MemwriteStatus() {
            return Fpga.FpgaRead(Fpga.memcontrol, 1);
        }
        
        // メモリー書き込み設定
        //   7 6 5 4 3 2 1 0
        //   |     | | | | +-- RUNMODE        0:即開始  1:ECに応じたメモリー長読み込んだ後TRGを待つ
        //   |     | | +-+---- EC(トリガー点) 0:先頭 1:1/8  2:中央  3:7/8
        //   |     +-+-------- SEL(入力切替)  0:CIC(adc1610)  1:LogicInput
        //   +---------------- D32(上位8bit)  0:下位32bit  1:上位8bit
        public static void MemwriteSetting( int sel ) {
            MemwriteSetting( sel, 0, 0 );
        }
        public static void MemwriteSetting( int sel, int endcount, int runmode) {
            Fpga.FpgaWrite(Fpga.memcontrol+1, (sel<<3)+(endcount<<1)+runmode);
        }
        
        // メモリーから全データ（buffsize分）読み出し
        public static int MemoryRead(Byte[] RBuf) {
            return MemoryRead( (int)buffsize, RBuf );
        }
        // メモリーからデータ（RBytes）読み出し
        public static int MemoryRead(int RBytes, Byte[] RBuf) {
            return Fpga.FpgaRead( Fpga.memory, RBytes, RBuf );
        }
        
        // メモリーから測定が終了したら全データ（buffsize分）読み出し
        public static int MemoryReadready(Byte[] RBuf) {
            while( true ) {
                int data = Fpga.FpgaRead( Fpga.memcontrol, 1 );
                if( data < 0 ) return -1;
                if( (data & 0x01) == 0 ) break;
                Thread.Sleep(1);
            }
            return MemoryRead( RBuf);
        }
        
        
        
        
        // ＡＤ９７０７に １バイト、WData を書く
        public static void ad9707Write( int WData ) {
            Fpga.FpgaWrite(Fpga.ad9707, WData );  //
        }
        // ＡＤ９７０７に WBytes（１～４）バイト、WData を書く
        public static void ad9707Write( int WBytes, int WData ) {
            Fpga.FpgaWrite(Fpga.ad9707, WBytes, WData );  //
        }
        
        
        
        
        // ＣＩＣに １バイト、WData を書く
        public static void CICWrite( int WData ) {
            Fpga.FpgaWrite(Fpga.cic, WData);
        }
        // ＣＩＣに WBytes（１～４）バイト、WData を書く
        public static void CICWrite( int WBytes, int WData ) {
            Fpga.FpgaWrite(Fpga.cic, WBytes, WData);
        }
        
        
        
        // UDA1345 の初期設定
        public static void uda1345_Init() {
            Fpga.FpgaWrite( Fpga.uda1345+1, 0x21 );  // 00 10 000 1 = 256fs I2S DCfiltering
            Thread.Sleep(10);
            Fpga.FpgaWrite( Fpga.uda1345+2, 0x00 );  // 00 000000   = Volume 0dB
            Thread.Sleep(10);
            Fpga.FpgaWrite( Fpga.uda1345+2, 0x80 );  // 100 00 0 00 = NoDeEmphasis NoMuting
            Thread.Sleep(10);
            Fpga.FpgaWrite( Fpga.uda1345+2, 0xC3 );  // 110000 11   = ADCon DACon
        }
        
        
        // UDA1345 の出力選択（ 0:off 1:DDS2 2:DDS3 3:DEMOD 4:I/Q 5:任意 6:オーディオ入力 7:固定パターン）
        public static void uda1345_OutSelect( int lch, int rch ) {
            Fpga.FpgaWrite( Apb3.Fpga.uda1345, (lch<<4)+rch );
        }
        
        
        
        
        // DEMOD 設定
        public static void demod_setting( int gain, int sel ) {
            Fpga.FpgaWrite( Apb3.Fpga.demod, (gain<<4)+sel );
        }
        
        
        // DEMOD レベル読み込み
        public static int demod_readlevel() {
            int level = Fpga.FpgaRead( Apb3.Fpga.demod );
            if( level<0 ) return level; // Read Error
            return (int)( Math.Sqrt( ((level&0xF0)>>4)*((level&0xF0)>>4) + (level&0x0F)*(level&0x0F) ) + 0.5);
        }
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        //================================================================================
        //
        //      ＦＴ２３２Ｈ 制御クラス
        //
        //================================================================================
        public static class Ft232h {
            
            private static FTDI ftdi = new FTDI();        // 
            private static uint ftdiDeviceCount = 0;      // 
            private static FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            
            public const Byte FtReadCmd  = 0x5C;  // Read
            public const Byte FtWriteCmd = 0x54;  // Write
            
            //
            //  オープン
            //
            public static bool FtOpen()
            {
                if (!isft232hopen)
                {
                    ftStatus = ftdi.GetNumberOfDevices(ref ftdiDeviceCount);
                    if ((ftStatus != FTDI.FT_STATUS.FT_OK) || (ftdiDeviceCount == 0))
                    {
                        mainf.LogPrint("\nApb3.Ft232h.FtOpen: ERROR  status=" + ftStatus.ToString() + ",  FTDI_device_count=" + ftdiDeviceCount.ToString());
                        return false;
                    }
                    
                    FTDI.FT_DEVICE_INFO_NODE[] ftdiDeviceList = new FTDI.FT_DEVICE_INFO_NODE[ftdiDeviceCount];
                    ftStatus = ftdi.GetDeviceList(ftdiDeviceList);
                    
                    for( int i=0; i<ftdiDeviceCount; i++ ) {
                        if ((ftdiDeviceList[i].Flags & 0x01) == 0) { // まだオープンしていない
                            string description  = ftdiDeviceList[i].Description.ToString();
                            if( (description.IndexOf("APB-3") == 0) && ( description.IndexOf("Ojisankoubou") > 0 ) ) {
                                ftStatus = ftdi.OpenByLocation(ftdiDeviceList[i].LocId);
                                if (ftStatus != FTDI.FT_STATUS.FT_OK) {
                                    mainf.LogPrint("\nApb3.Ft232h.FtOpen: ERROR  Failed to open device (" + ftStatus.ToString() + ")");
                                    return false;
                                }
                                ftdi.SetTimeouts(1000, 1000); // ReadTimeout=1000ms, WriteTimeout=1000ms
                                ftdi.SetBaudRate(3000000);  // Baudrate設定 3Mbps以下
                                ftdi.SetLatency(2);         // Latency設定 2-255ms
                                
                                isft232hopen = true;
                                mainf.LogPrint("\nApb3.Ft232h.FtOpen: DeviceIndex=" + i.ToString() + "  LocationID=" + ftdiDeviceList[i].LocId + "  Description=" + description);
                                return true;
                            }
                        }
                    }
                    mainf.LogPrint("\nApb3.Ft232h.FtOpen: ERROR  No APB-3 board is found");
                    return false;
                }
                return true;
            }
            
            
            //
            //  クローズ
            //
            public static void FtClose()
            {
                ftdi.Close(); // 終了時 FT232H を閉じる
                isft232hopen = false;
                mainf.LogPrint("\nApb3.Ft232h.FtClose: Connection closed");
            }
            
            
            //
            //  接続されている FTDI デバイスの情報
            //
            public static string FtInfo()
            {
                String str = "\n";
                ftStatus = ftdi.GetNumberOfDevices(ref ftdiDeviceCount);
                if ((ftStatus == FTDI.FT_STATUS.FT_OK) && (ftdiDeviceCount != 0))
                {
                    str += "\nNumber of FTDI devices: " + ftdiDeviceCount.ToString();
                }
                else
                {
                    str += "\nApb3.Ft232h.FtInfo: No FTDI device is detected";
                    return str;
                }
                FTDI.FT_DEVICE_INFO_NODE[] ftdiDeviceList = new FTDI.FT_DEVICE_INFO_NODE[ftdiDeviceCount];
                ftStatus = ftdi.GetDeviceList(ftdiDeviceList);
                if (ftStatus == FTDI.FT_STATUS.FT_OK)
                {
                    for (UInt32 i = 0; i < ftdiDeviceCount; i++)
                    {
                        str += "\nDevice Index = " + i.ToString();
                        str += "\n  Flags: " + String.Format("{0:x}", ftdiDeviceList[i].Flags);
                        if ((ftdiDeviceList[i].Flags & 0x02) == 0) str += " ,12Mbps"; else str += " ,480Mbps";
                        if ((ftdiDeviceList[i].Flags & 0x01) == 0) str += " ,未使用"; else str += " ,使用中";
                        str += "\n  Type: " + ftdiDeviceList[i].Type.ToString();
                        str += "\n  ID: " + ftdiDeviceList[i].ID.ToString();
                        str += "\n  Location ID: " + ftdiDeviceList[i].LocId.ToString();
                        str += "\n  Serial Number: " + ftdiDeviceList[i].SerialNumber.ToString();
                        str += "\n  Description: " + ftdiDeviceList[i].Description.ToString();
                    }
                }
                return str;
            }
            
            
            //
            //  書き込み
            //
            public static int FtWrite( byte[] Wbuf, int size )
            {
                uint BytesWritten = 0;
                ftStatus = ftdi.Write(Wbuf, (int)size, ref BytesWritten);
                if( (ftStatus != FTDI.FT_STATUS.FT_OK) || (BytesWritten != size) ) {
                    mainf.LogPrint( "\nApb3.Ft232h.FtWrite: ERROR  ftStatus=" + ftStatus.ToString() + "  BytesWritten=" + BytesWritten.ToString() + " (" + size.ToString() + ")" );
                }
                return (int)BytesWritten;
            }
            //
            //  書き込み　1～4 バイト
            //
            public static int FtWrite( int data, int size )
            {
                byte[] Wbuf = new Byte[4];
                switch( size ) {
                case 1:
                    Wbuf[0] = (byte)(data);
                    break;
                case 2:
                    Wbuf[0] = (byte)(data >> 8);
                    Wbuf[1] = (byte)(data);
                    break;
                case 3:
                    Wbuf[0] = (byte)(data >> 16);
                    Wbuf[1] = (byte)(data >> 8);
                    Wbuf[2] = (byte)(data);
                    break;
                default:
                    Wbuf[0] = (byte)(data >> 24);
                    Wbuf[1] = (byte)(data >> 16);
                    Wbuf[2] = (byte)(data >> 8);
                    Wbuf[3] = (byte)(data);
                    break;
                }
                return FtWrite( Wbuf, size );
            }
            //
            //  書き込み　1バイト
            //
            public static int FtWrite( int data )
            {
                return FtWrite( data, 1 );
            }
            
            
            //
            //  読み出し
            //
            public static int FtRead( byte[] Rbuf, int size )
            {
                uint BytesRead = 0;
                ftStatus = ftdi.Read( Rbuf, (uint)size, ref BytesRead );
                if( (ftStatus != FTDI.FT_STATUS.FT_OK) || (BytesRead != size) ) {
                    mainf.LogPrint( "\nApb3.Ft232h.FtRead: ERROR  ftStatus=" + ftStatus.ToString() + "  BytesRead=" + BytesRead.ToString() + " (" + size.ToString() + ")" );
                }
                return (int)BytesRead;
            }
            
            
            //
            //  ASYNC BIT BANG モード   mask: ポート入出力設定(0:入力 1:出力)
            //
            public static void FtAsyncBitBang( byte mask )
            {
                ftStatus = ftdi.SetBitMode(mask, FTDI.FT_BIT_MODES.FT_BIT_MODE_ASYNC_BITBANG);
            }
            
            
            //
            //  SYNC BIT BANG モード   mask: ポート入出力設定(0:入力 1:出力)
            //
            public static void FtSyncBitBang( byte mask )
            {
                ftdi.SetBitMode(mask, FTDI.FT_BIT_MODES.FT_BIT_MODE_SYNC_BITBANG);
            }
            
            
            //
            //  C BUS BITBANG モード    mask: bit7..4 はポート入出力設定(0:入力 1:出力)、bit3..0 は出力ポートの値
            //
            public static void FtCbusBitBang( byte mask )
            {
                ftdi.SetBitMode(mask, FTDI.FT_BIT_MODES.FT_BIT_MODE_CBUS_BITBANG);
            }
            
            
            //
            //  BitBang モードをリセット
            //
            public static void FtBitModeReset()
            {
                ftdi.SetBitMode(0xFF, FTDI.FT_BIT_MODES.FT_BIT_MODE_RESET);
            }
            
            
            //
            //  BitBang モードでのポート入力   CBUS BITBANG では下位4ビットに入る
            //
            public static byte FtGetBit()
            {
                byte data=0;
                ftdi.GetPinStates(ref data);    // FT_GetBitMode
                return data;
            }
            
            
            //
            //  RX/TXバッファをクリアする
            //
            public static void FtPurge()
            {
                ftdi.Purge(FTDI.FT_PURGE.FT_PURGE_RX | FTDI.FT_PURGE.FT_PURGE_TX);
            }
            
            
            //
            //  Baudrate 設定
            //
            public static void FtSetBaudRate( int baudrate )
            {
                ftdi.SetBaudRate( (uint)baudrate );
            }
            
            
            
        }  // class Ft232h
        
        
        
        
        
        
        
        
        //================================================================================
        //
        //      ＦＰＧＡ 制御クラス
        //
        //================================================================================
        public static class Fpga {
            public const  double clockfreq  = 100e6;                      // FPGA 水晶発振周波数
            public const int buffsize   = 8192*4; // FPGA のバッファーサイズ  8192words * 40bits
            
            // FPGA の各モジュールアドレス
            public const int signature = 0x0000; // R: APB3....date(4)
            public const int memory    = 0x8000; // 32KB
            public const int dds       = 0x0100; // W: FREQ(31..0)  little endian
            public const int dds0      = dds   ; // W: FREQ(31..0)  little endian
            public const int dds1      = dds+8 ; // W: FREQ(31..0)  little endian
            public const int dds2      = dds+16; // W: FREQ(31..0)  little endian
            public const int dds3      = dds+24; // W: FREQ(31..0)  little endian
            public const int cic       = 0x0200; // W: decimation rate(3..0)  0..12:1,2,4,8,16,32,64,128,256,512,1024,2048,4096  
            public const int adc1610   = 0x0300; // W:   R:
            public const int ad9707    = 0x0400; // W: 1:enable  0:dds select
//            public const  int logiana   = 0x0500; // W: 00   3:DBL 2:AQMODE 1:EDGE 0:SLOPE
//                                                  //    01   4..0: TIMEBASE
//                                                  //    02-3 TRGMASK
//                                                  //    04-5 TRGMATCH
            public const int uda1345   = 0x0600;  // W: 600 6..4:SelLch  2..0 SelRch
                                                  //    601 L3 出力データ(STATUS)
                                                  //    602 L3 出力データ(DATA)
            
            public const int memcontrol= 0x0700;  // W: 700 bit0 に 1 を書くと書き込み開始、0 を書くと中断
                                                  //    701 7:D32上位８ビット読み出し 3:入力SEL 2..1:EC（トリガー位置） 0:RUNMODE(0:FREERUN 1:TRG)
                                                  // R: 700 2:TRGD  1:TRGWAIT  0:START
                                                  //    701 ADRCOUNT(7 downto 0)
                                                  //    702 "000" & ADRCOUNT(12 downto 8);
                                                  
            public const int freqdet = 0x0800;    // W: 800   7 6 5 4 3 2 1 0
                                                  //                        +-- START/STOP  0 を書くと測定停止、1 を書くと周波数カウンタ測定開始　
                                                  //  
                                                  //  
                                                  //    801   7 6 5 4 3 2 1 0
                                                  //            | | | | +-+-+-- MESBITSEL   0:1/1  1:1/2    2:1/4   ...  7:1/128　
                                                  //            | | +-+-------- PRESCALER   0:1/1  1:1/256  2:1/65536
                                                  //            | +------------ FCINSEL     周波数カウンター入力切替  0:直接 1:周波数変換後
                                                  //            +-------------- SEQ         0:単発測定   1: 連続測定
                                                  //  
                                                  //    802   7 6 5 4 3 2 1 0
                                                  //          | | | | +-+-+-+-- SIGSEL     入力切替 0:D0 1:D1 ... 14:D14 15:アナログ入力
                                                  //          | +-+-+---------- FMCICSEL   FM 検波器 CIC 段数切替  0:128 1:256 ... 5:4096
                                                  //          +---------------- TRGEDGE    トリガー出力　0:立ち上がり　1:立下り
                                                  //
                                                  //    803   7 6 5 4 3 2 1 0
                                                  //                   +-+-+-+-- FMSEQN     何回ごとに記録するか 0:1回　1:2回　2:4回  ... 15:32768回
                                                  //                                        PHASEOV & PHASE
                                                  //
                                                  // R: 800   7 6 5 4 3 2 1 0
                                                  //                    | +-+-- STATE  0:測定終了 1:測定開始 2:開始カウンター値取得 or 連続測定中
                                                  //                    +------ FSIGN  周波数の符号 0:正 1:負
                                                  //  
                                                  //    801   SEQCNT   (  7 downto  0 )
                                                  //    802   MESCNT   (  7 downto  0 )
                                                  //    803   CNT_STOP (  7 downto  0 )
                                                  //    804   CNT_STOP ( 15 downto  8 )
                                                  //    805   CNT_STOP ( 23 downto 16 )
                                                  //    806   CNT_STOP ( 31 downto 24 )
                                                  //    807   CNT_START(  7 downto  0 )
                                                  //    808   CNT_START( 15 downto  8 )
                                                  //    809   CNT_START( 23 downto 16 )
                                                  //    80A   CNT_START( 31 downto 24 )
            
            public const int demod= 0x0A00;       // W: 00 7..4:gain 1..0:sel(0:LSB 1:USB 2:AM 3:I+Q)
            
            public const int oscilloscope=0x0B00; // W: 7..6:trgmode 2..0:threshold
            
            
            
            //
            //  FPGA config
            //
            public static bool FpgaConfig()
            {
                isconfigured = false;
                mainf.LogPrint("\n\nFpga.FpgaConfig");
                if( !Ft232h.FtOpen() ) return false;
                    Ft232h.FtCbusBitBang(0x88);   // PROG_B(CBUS3) = H
                    Thread.Sleep(1);
                    
                    Ft232h.FtCbusBitBang(0x80);   // PROG_B(CBUS3) = L
                    Thread.Sleep(1);
                    
                    Ft232h.FtCbusBitBang(0x88);   // PROG_B(CBUS3) = H
                    Thread.Sleep(1);            // INIT_B 待ち
                    
                    Thread.Sleep(100);  // M.SPI なら 50ms でコンフィグされるので待つ
                    byte done = Ft232h.FtGetBit();
                    if( (done&0x04) != 0 ) {  // Done==1 ?
                        mainf.LogPrint("\n M.SPI configured");
                    }else{              // コンフィグされていなかったら S.S.MAP モード
                        OpenConfigFile();
                        if (filesize != 0){
                            sw.Restart();
                            
                            Ft232h.FtAsyncBitBang(0xFF);  // ASYNC BITBANG モード　全ビット出力
                            int BytesWritten = Ft232h.FtWrite(filebuf, filesize + 1000);  // 追加CCLK=1000
                            mainf.LogPrint("\n S.S.MAP configured  実行時間=", sw);
                        }else{
                            mainf.LogPrint("\n ERROR: OpenFile failed");
                        }
                    }
                    
                    Ft232h.FtCbusBitBang(0x88);   // PROG_B(CBUS3) = H
                    done = Ft232h.FtGetBit();
                    if( (done&0x04) == 0 ) {    // Done は bit2
                        mainf.LogPrint("\n ERROR: FpgaConfig failed" );
                    }else{
                        isconfigured = true;
                    }
                    
                    Ft232h.FtBitModeReset();      // BitBangMode をリセットする
                    Ft232h.FtPurge();             // RX/TXバッファをクリアする
                    
                    Thread.Sleep(100);          // このあと再度バッファクリアするが 10 以下にすると RX バッファにデータが残る
                    Ft232h.FtPurge();             // RX/TXバッファをクリアする
                    
                return isconfigured;
            }
            
            
            //
            //  FPGA の内部モジュールから読み出し( １バイト ）
            //
            public static int FpgaRead( int fpgaadrs )
            {
                return FpgaRead( fpgaadrs, 1 );
            }
            
            //
            //  FPGA の内部モジュールから読み出し( size は １～３ ）
            //    読んだデータを返す　エラーのときは負数
            //
            public static int FpgaRead( int fpgaadrs, int size )
            {
                if( size > 3  ) return error.invalid;
                
                Byte[] Wbuf = new Byte[5];
                Wbuf[0] = Ft232h.FtReadCmd; //0x5C;         // Read cmd
                Wbuf[1] = (byte)(fpgaadrs);
                Wbuf[2] = (byte)(fpgaadrs >> 8);
                Wbuf[3] = (byte)size;
                Wbuf[4] = 0x00;
                
                Ft232h.FtWrite(Wbuf, 5);
                
                Byte[] Rbuf = new Byte[4];
                if( Ft232h.FtRead(Rbuf, size) == size ) {
                  int data=0;
                  for( int i=size-1; i>=0; i-- ) data = ((int)data<<8) + (int)Rbuf[i];
                  return data;
                }else{
                  return error.err;
                }
            }
            
            //
            //  FPGA の内部モジュールから読み出し( size は １～３２７６８ ）
            //    読んだバイト数を返す　エラーのときは負数
            //
            public static int FpgaRead( int fpgaadrs, int size, byte[] buf )
            {
                if( size > 32768 ) return error.invalid;
                
                try {
                    Byte[] Wbuf = new Byte[5];
                    Wbuf[0] = Ft232h.FtReadCmd; //0x5C;         // Read cmd
                    Wbuf[1] = (byte)(fpgaadrs);
                    Wbuf[2] = (byte)(fpgaadrs >> 8);
                    Wbuf[3] = (byte)(size);
                    Wbuf[4] = (byte)(size >> 8);
                    
                    Ft232h.FtWrite(Wbuf, 5);
                    
                    int numBytesRead = (int)Ft232h.FtRead(buf, size);
                    return ( ( numBytesRead == size )? numBytesRead : error.err);
                }
                catch (Exception ex) {
                    mainf.LogPrint("\nFpga.Read: e="+ ex.Message);
                    return error.err;
                }
            }
            
            
            //
            //  FPGA の内部モジュールへの書き込み(１バイト ）
            //
            public static int FpgaWrite( int fpgaadrs, int data )
            {
                return FpgaWrite( fpgaadrs, 1, data );
            }
            
            //  FPGA の内部モジュールへの書き込み( size は １～４ ）
            public static int FpgaWrite( int fpgaadrs, int size, int data )
            {
                if( size>4  ) return error.invalid;
                
                try {
                    Byte[] Wbuf = new Byte[5+4];
                    Wbuf[0] = Ft232h.FtWriteCmd;  //0x54;         // Write cmd
                    Wbuf[1] = (byte)(fpgaadrs);
                    Wbuf[2] = (byte)(fpgaadrs >> 8);
                    Wbuf[3] = (byte)size;
                    Wbuf[4] = 0x00;
                    
                    for (int i = 0; i < 4; i++) {
                        Wbuf[i+5] = (Byte)(data);      // set write data
                        data = data >> 8;
                    }
                    
                    int ByteWritten = (int)Ft232h.FtWrite(Wbuf, (int)(5+size)) - 5;
                    return (( ByteWritten == size )? ByteWritten: -1);
                }
                catch (Exception ex) {
                    mainf.LogPrint("\nFpga.Write: e="+ ex.Message );
                    return error.err;
                }
            }
            
            
            
        } // Class Fpga
        
        
        
        
        
        
        
        
        
        
        //================================================================================
        //
        //      ＤａｔａＦｌａｓｈ (AT25DF041) 制御クラス
        //
        //================================================================================
        public static class Df {
            //
            //  DataFlash から読む
            //
            public static void DfRead( int dfadrs, int size, byte[] buf )
            {
                mainf.LogPrint("\nDf.Read dfadrs=" + dfadrs.ToString("X6") + " size=" + size.ToString("X6") );
                
                int remain=size;
                int offset=0;
                while( remain != 0 ) {
                    size = ( remain > 256 )? 256: remain;
                    DfRead256( dfadrs, size, buf, offset );
                    remain -= size;
                    dfadrs += size;
                    offset += size;
                }
            }
            //
            //  DataFlash から読む（ size は 1～256 ）
            //
            private static void DfRead256( int dfadrs, int size, byte[] buf, int offset )
            {
                tmpbuf[0] = 0x03;           // 0x03=Read Array
                for( int i=1; i<4; i++ ) {
                    tmpbuf[i] = (byte)(dfadrs>>16);
                    dfadrs = dfadrs<<8;
                }
                DfWriteRead( 4+size, tmpbuf );      // バッファに op+adrs を書く
                
                DfBufRead( 4+size, tmpbuf );  // バッファに DF から読んだデータが入るのでバッファから読む
                Buffer.BlockCopy( tmpbuf, 4, buf, (int)offset, (int)size ); // 5 バイト目以降が有効なデータ
                
                return;
            }
            
            
            //
            //  DataFlash に書く
            //
            public static int DfWrite( int dfadrs, int size, byte[] buf )
            {
                mainf.LogPrint("\nDf.Write dfadrs=" + dfadrs.ToString("X6") + " size=" + size.ToString("X6") );
                
                sw.Restart();
                int remain=size;
                int offset=0;
                while( remain != 0 ) {
                    if( (dfadrs & 0x000000FF) != 0 ) {      // 256バイト境界に合わせる
                        size = 256 - (dfadrs & 0x000000FF);
                    }else{
                        size = ( remain > 256 )? 256: remain;
                    }
                    DfWrite256( dfadrs, size, buf, offset );
                    remain -= size;
                    dfadrs += size;
                    offset += size;
                }
                return 0;
            }
            //
            //  DataFlash に書く（ size は 1～256 ）
            //
            private static int DfWrite256( int dfadrs, int size, byte[] buf, int offset )
            {
                if( !DfBusyWait() ) return -1;  // timeout
                DfSetWEL();               // WEL(WriteEnable) ビットをセットする
                
                tmpbuf[0] = 0x02;           // 0x02=Byte/Page Program
                for( int i=1; i<4; i++ ) {
                    tmpbuf[i] = (byte)(dfadrs>>16);
                    dfadrs = dfadrs<<8;
                }
                Buffer.BlockCopy( buf, (int)offset, tmpbuf, 4, (int)size ); // 5 バイト目以降にデータコピー
                DfWriteRead( 4+size, tmpbuf ); // op+adrs+data を書く
                return 0;
            }
            
            
            //
            //  指定セクターをアンプロテクトする
            //
            public static void DfUnProtectSector( int dfadrs )
            {
                mainf.LogPrint("\nDf.UnProtect dfadrs=" + dfadrs.ToString("X6") );
                
                DfSetWEL();               // WEL(WriteEnable) ビットをセットする
                
                tmpbuf[0] = 0x39;       // 0x39=Unprotect Sector
                for( int i=1; i<4; i++ ) {
                    tmpbuf[i] = (byte)(dfadrs>>16);
                    dfadrs = dfadrs<<8;
                }
                DfWriteRead( 4, tmpbuf );
            }
            
            
            //
            //  全セクターをアンプロテクトする
            //
            public static void DfUnProtectAll()
            {
                mainf.LogPrint("\nDf.UnProtectAll" );
                
                DfSetWEL();               // WEL(WriteEnable) ビットをセットする
                
                byte[] buf = new byte[2]{ 0x01, 0x00 }; // 0x01=Write Status Register
                DfWriteRead( 2, buf );
            }
            
            
            //
            //  全セクターをプロテクトする
            //
            public static void DfProtectAll()
            {
                mainf.LogPrint("\nDf.ProtectAll" );
                
                DfSetWEL();               // WEL(WriteEnable) ビットをセットする
                
                byte[] buf = new byte[2]{ 0x01, 0xFF }; // 0x01=Write Status Register
                DfWriteRead( 2, buf );
            }
            
            
            //
            //  ブロック（4KB）イレーズする
            //
            public static int DfBlockErase4k( int dfadrs )
            {
                mainf.LogPrint("\nDf.BlockErase4k dfadrs=" + dfadrs.ToString("X6") );
                
                return DfBlockErase( 0x20, dfadrs );
            }
            //
            //  ブロック（32KB）イレーズする
            //
            public static int DfBlockErase32k( int dfadrs )
            {
                mainf.LogPrint("\nDf.BlockErase32k dfadrs=" + dfadrs.ToString("X6") );
                
                return DfBlockErase( 0x52, dfadrs );
            }
            //
            //  ブロック（64KB）イレーズする
            //
            public static int DfBlockErase64k( int dfadrs )
            {
                mainf.LogPrint("\nDf.BlockErase64k dfadrs=" + dfadrs.ToString("X6") );
                
                return DfBlockErase( 0xD8, dfadrs );
            }
            //
            //  ブロックイレーズする
            //
            private static int DfBlockErase( byte op, int dfadrs )
            {
                DfUnProtectAll();
                
                if( !DfBusyWait() ) return -1;  // timeout
                DfSetWEL();               // WEL(WriteEnable) ビットをセットする
                
                tmpbuf[0] = op;         // Byte erase op 4k:0x20 32k:0x52 64k:0xD8
                for( int i=1; i<4; i++ ) {
                    tmpbuf[i] = (byte)(dfadrs>>16);
                    dfadrs = dfadrs<<8;
                }
                DfWriteRead( 4, tmpbuf );
                if( !DfBusyWait(1000) ) return -1;  // timeout 1秒
                return 0;
            }
            
            
            //
            //  チップ（全メモリーエリア）イレーズする
            //
            public static int DfChipErase()
            {
                mainf.LogPrint("\nDf.ChipErase" );
                
                DfUnProtectAll();
                
                if( !DfBusyWait() ) return -1;  //
                DfSetWEL();               // WEL(WriteEnable) ビットをセットする
                
                tmpbuf[0] = 0x60;       // 0x60=Chip Erase
                DfWriteRead( 1, tmpbuf );
                
                if( !DfBusyWait(10000) ) return -1;  // timeout 10秒
                return 0;
            }
            
            
            //
            //  DataFlash の ID を読む
            //
            public static string DfReadID()
            {
                mainf.LogPrint("\nDf.ReadID" );
                
                tmpbuf[0] = 0x9F;       // 0x9F=Read ID
                DfWriteRead( 5, tmpbuf );
                DfBufRead( 5, tmpbuf );
                string str = "";
                for( int i=1; i<5; i++ ) str += tmpbuf[i].ToString("X2");
                str += ( tmpbuf[1]==0x1F )? " (ATMEL)":" ???"; // Manufacturer ID
                return str;
            }
            
            
            //
            //  WEL(Write Enable Latch) ビットをセットする
            //
            private static void DfSetWEL()
            {
                tmpbuf[0] = 0x06;       // 0x06=Write Enable
                DfWriteRead( 1, tmpbuf );
            }
            
            
            //
            //  ビジーならタイムアウトするまで待つ ( false:timeout )
            //
            private static bool DfBusyWait()
            {
                return DfBusyWait( 100 ); // default = 100ms
            }
            private static bool DfBusyWait( int timeout )
            {
                int t=timeout;
                while( DfIsBusy() ) {
                    Thread.Sleep(1);
                    if( (timeout--)==0 ) {
                        mainf.LogPrint("\nBusyWait: ERROR Timeout=" + t.ToString() + "ms");
                        return false;  // timeout
                    }
                }
                return true;
            }
            
            
            //
            //  ビジー？ ( true:Busy  false:Ready )
            //
            private static bool DfIsBusy()
            {
                byte status = DfReadStatus(); // bit0 = Busy
                return ((status&0x01)==1);
            }
            
            
            //
            //  ステータスレジスターを読む
            //  (7:SPRL 6:SPM 5:EPE 4:WPP 3..2:SWP 1:WEL 0:RDY/BSY)
            private static byte DfReadStatus()
            {
                tmpbuf[0] = 0x05;       // 0x05=Read Status Register
                DfWriteRead( 2, tmpbuf );
                DfBufRead( 2, tmpbuf );
                return tmpbuf[1];
            }
            
            
            //
            //  DF 読み書き実行( DF に書くと同時に DF から読んだデータをバッファーに書く )
            //
            private static int DfWriteRead( int size, byte[] buf )
            {
                Byte[] Wbuf = new Byte[3+size];
                Wbuf[0] = 0x44;         // DF_RW
                Wbuf[1] = (byte)(size);
                Wbuf[2] = (byte)(size >> 8);
                Buffer.BlockCopy( buf, 0, Wbuf, 3, (int)size );
                return Ft232h.FtWrite(Wbuf, 3+size);  // BytesWritten を返す
            }
            
            
            //
            //  バッファー読み出し（BytesRead を返す）
            //
            private static int DfBufRead( int size, byte[] buf )
            {
                Byte[] Wbuf = new Byte[3];
                Wbuf[0] = 0x45;         // DF_BUF_READ
                Wbuf[1] = (byte)(size);
                Wbuf[2] = (byte)(size >> 8);
                Ft232h.FtWrite(Wbuf, 3);
                int BytesRead = Ft232h.FtRead(buf, size);
                return BytesRead;
            }
            
            
            //
            //  コンフィグファイルをデータフラッシュに書き込む
            //
            public static bool DfWriteConfig()
            {
                mainf.LogPrint( "\n\nApb3.DfWriteConfig:" );
                if (Apb3.Open())
                {
                    Apb3.OpenConfigFile();
                    if (Apb3.filesize != 0)
                    {
                        // bit ファイルの先頭の不要な部分を削除（Spartan6 ではこの部分があると SPI コンフィグ失敗する）
                        int offset;
                        for( offset=0; offset<256; offset++ ) {        // コンフィグ有効データ開始位置を探す
                            if( Apb3.filebuf[offset] == 0xFF ) break;
                        }
                        if( offset == 256 ) {  // コンフィグ有効データが見つからなかった
                            mainf.LogPrint("ERROR   Invalid config file");
                            Apb3.Close();
                            return false;
                        }
                        int size = 0;
                        int remain = Apb3.filesize-offset;
                        int dfadrs = 0;
                        byte[] buf = new byte[32768];
                        DfChipErase();
                        DfUnProtectAll();
                        while( remain != 0 ) {
//                            Print( "." );
//                            Thread.Sleep(1);
                            size = ( remain >= 32768 )? 32768: remain;
                            Buffer.BlockCopy( Apb3.filebuf, (int)offset, buf, 0, (int)size ); // データコピー
                            DfWrite( dfadrs, size, buf );
                            remain -= size;
                            offset += size;
                            dfadrs += size;
                        }
                        mainf.LogPrint( "\nDone" );
                        DfProtectAll();
                    }else{
                        mainf.LogPrint("ERROR  OpenFile failed");
                        Apb3.Close();
                        return false;
                    }
                    Apb3.Close();
                    return true;
                }
                return false;
            }
            
            
            
        }  // class Df
        
        
        
        
        
        
        
        //================================================================================
        //
        //      ＡＤＣ (ADC1610S) 制御クラス
        //
        //================================================================================
        public static class Adc1610 {
            public static int[] reg  = new int[13]{
                0x05, 0x06, 0x08, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x20, 0x21, 0x22
            };  // ADC1610S のレジスターアドレス
            
            public static int[] reginit = new int[13]{
                0x00, 0x01, 0x0E, 0x11, 0x09, 0x00, 0x00, 0xA5, 0x5A, 0x0F, 0x0E, 0x00, 0x00
            };  // ADC1610S のレジスター初期値　仕様書 p.30参照
            
            
            //
            //  ADC1610S のレジスタ初期化と遅延自動設定   true:Success  false:Fail
            //
            public static bool AdcInit()
            {
                for( int i=0; i<13; i++ ) {        // レジスター初期設定
                    AdcRegWrite( reg[i], reginit[i] );
                }
                if( !AdcIoDelayCal() ) return false;   // 遅延自動設定
                return true;
            }
            
            
            //
            //  ADC1610S のレジスタ読み出し
            //
            public static int AdcRegRead( int regadrs )
            {
                Fpga.FpgaWrite( Fpga.adc1610+1, 1, 0x80 + regadrs);   // Read(bit7=1) + REG_ADRS を設定
                Fpga.FpgaWrite( Fpga.adc1610, 1, 1 );           // adc1610 と SPI 通信開始
                return Fpga.FpgaRead( Fpga.adc1610, 1 );        // adc1610 から読み込んだデータ
            }
            
            
            //
            //  ADC1610S のレジスタ書き込み
            //
            public static int AdcRegWrite( int regadrs, int data )
            {
                if( Fpga.FpgaWrite( Fpga.adc1610+1, 1, regadrs) < 0 )  return -1; // Write(bit7=0) + REG_ADRS を設定
                Fpga.FpgaWrite( Fpga.adc1610+2, 1, data );      // adc1610 に書き込むデータ設定
                Fpga.FpgaWrite( Fpga.adc1610, 1, 1 );           // adc1610 と SPI 通信開始
                return 0;
            }
            
            
            //
            //  ADC1610S データ入力の IODELAY2 遅延を自動設定   true:Success  false:Fail
            //
            public static bool AdcIoDelayCal()
            {
                adccaldly = -1;
                AdcRegWrite( 0x05, 0x00 );                 // RegAdrs=5 opmode=0(Power-up)
                AdcRegWrite( 0x14, 0x07 );                 // テストパターン
                Fpga.FpgaWrite( Apb3.Fpga.adc1610 + 4, 1, 1 );  //遅延ごとのエラーテーブル作成開始
                Thread.Sleep(10);
                int dlytbl = Fpga.FpgaRead( Apb3.Fpga.adc1610 + 4, 2 );  //エラーテーブル
                mainf.LogPrint("\n\nIoDelayCal: dlytbl=" + dlytbl.ToString("X4"));
                
                int[] okstarttbl = new int[2]{ -1, 0 };
                int[] okwidthtbl = new int[2]{ -1, 0 };
                int okcnt=0;
                for( int i=0; i<16; i++ ) {
                    if( (dlytbl&0x0001) == 0 ) {    // OK
                        if( okwidthtbl[okcnt] == 0 ) {
                            okstarttbl[okcnt] = i;
                        }
                        okwidthtbl[okcnt]++;
                    }else{                          // NG
                        if( okwidthtbl[okcnt] != 0 ) {
                            okcnt++;
                            if( okcnt == 2 ) {  // OK 領域が3つ以上ある
                                mainf.LogPrint("\n CAL ERROR: 遅延 OK 領域が3つ以上ある");
                                return false; // OK になる範囲が細かく分かれすぎている
                            }
                        }
                    }
                    dlytbl = dlytbl>>1;
                }
                int okwidth = 0;
                int dly = 0;
                if (okwidthtbl[0] > okwidthtbl[1]) {
                    okwidth = okwidthtbl[0];
                    dly = (okstarttbl[0]<<4) + (okwidth<<3);
                }else{
                    okwidth = okwidthtbl[1];
                    dly = (okstarttbl[1]<<4) + (okwidth<<3);
                }
                mainf.LogPrint(", dly = " + dly.ToString("X2"));
                if(okwidth < 3) {
                    mainf.LogPrint("\n CAL ERROR: 遅延 OK の範囲が 2 以下");
                    return false; // OK になる遅延値の範囲が狭すぎる
                }
                
                AdcIoDelaySetting( (int)dly );
                AdcRegWrite( 0x14, 0 );                // テストパターン → ノーマルモードへ
                return true;
            }
            
            
            //
            //  ADC1610S データ入力の IODELAY2 遅延設定
            //
            public static void AdcIoDelaySetting( int dly )
            {
                adccaldly = dly;
                AdcIoDelaySetting();
            }
            public static void AdcIoDelaySetting()
            {
                Fpga.FpgaWrite( Fpga.adc1610+5, 1, adccaldly );
            }
            
            
            //
            //  ADC1610S 再設定
            //
            public static int AdcReInit()
            {
                for( int i=0; i<13; i++ ) {        // レジスター初期設定
                    if( AdcRegWrite( reg[i], reginit[i] ) < 0 ) return error.err;
                }
                if( adccaldly < 0 ) {
                    if( !AdcIoDelayCal() ) return error.err;
                }else{
                    AdcIoDelaySetting();                 // IO delay
                }
                return 0;
            }
            
            
            
            
        }  // class Adc1610
        
        
        
    }
}
